#define WNCK_I_KNOW_THIS_IS_UNSTABLE

/* This file is largely copied from the libwnck example code */

#include "config.h"

#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>

#include <libwnck/libwnck.h>
#include <gtk/gtk.h>
#include <libxml/tree.h>

#include "global.h"

#include <rox.h>

#include "window-menu.h"

char *app_dir = NULL;

static GdkWindow *socket = NULL;	/* NULL => Not an applet */

static WnckScreen *screen = NULL;

static GtkWidget *pager = NULL;

static GtkWidget *frame = NULL;

static GtkWidget *toplevel = NULL;


static ROXOption n_rows;

static ROXOption frame_type;

static int
workspace_at_point (WnckPager *pager,
                    int        x,
                    int        y,
                    int       *viewport_x,
                    int       *viewport_y);

/* Read the contents of this property. g_free() the result. */
static char *read_property(GdkWindow *window, GdkAtom prop,
			   GdkAtom type, gint *out_length)
{
	gint	grab_len = 4096;
	gint	length;
	guchar	*data;

	while (1)
	{
		if (!(gdk_property_get(window, prop, type, 0, grab_len,
				FALSE, NULL, NULL, &length, &data) && data))
			return NULL;	/* Error? */

		if (length >= grab_len)
		{
			/* Didn't get all of it - try again */
			grab_len <<= 1;
			g_free(data);
			continue;
		}

		*out_length = length;

		return data;
	}
}

static char *read_string(GdkWindow *window, GdkAtom prop)
{
	gint  len;
	gchar *data;

	data = read_property(window, prop,
			gdk_atom_intern("STRING", FALSE), &len);
	
	if (!data)
		return NULL;

	data = g_realloc(data, len + 1);
	data[len] = '\0';
	return data;
}

static void position(GtkMenu *menu, gint *x, gint *y, gboolean *push_in,
		     gpointer user_data)
{
	GtkRequisition req;
	int  margin = 2;
	int screen_width, screen_height;
	char *pos = (char *) user_data;
	char *comma;

	*push_in = TRUE;

	gdk_window_get_pointer(NULL, x, y, NULL);

	screen_width = gdk_screen_width();
	screen_height = gdk_screen_height();

	if (pos)
	{
		comma = strchr(pos, ',');
		if (comma)
			margin = atoi(comma + 1);
	}

	gtk_widget_size_request(GTK_WIDGET(menu), &req);

	if (!pos)
	{
		*x -= req.width / 2;
		*y -= 32;
		goto out;
	}
	
	if (pos[0] == 'T')
	{
		*y = margin;
		*x -= req.width / 4;
	}
	else if (pos[0] == 'B')
	{
		*y = screen_height - margin - req.height;
		*x -= req.width / 4;
	}
	else if (pos[0] == 'L')
	{
		*x = margin;
		*y -= 16;
	}
	else
	{
		*x = screen_width - margin - req.width;
		*y -= 16;
	}

	g_free(pos);

out:
	*x = CLAMP(*x, 4, screen_width - 4 - req.width);
	*y = CLAMP(*y, 4, screen_height - 4 - req.height);
}

static void show_help(void)
{
	char *argv[3] = {"rox", NULL, NULL};
	GError *error = NULL;

	argv[1] = g_build_filename(getenv("APP_DIR"), "Help", NULL);
	g_spawn_async(NULL, argv, NULL, G_SPAWN_SEARCH_PATH,
			NULL, NULL, NULL, &error);
	g_free(argv[1]);

	if (error)
	{
		GtkWidget *box;
		box = gtk_message_dialog_new(NULL, 0, GTK_MESSAGE_ERROR,
					GTK_BUTTONS_OK, "%s", error->message);
		gtk_dialog_run(GTK_DIALOG(box));
		gtk_widget_destroy(box);
		g_error_free(error);
	}
}

static GList *get_window_list(void)
{		
	GList *window_list = NULL;
	GList *l = wnck_screen_get_windows_stacked(screen);	
	while (l)
	{
		WnckWindow *window = WNCK_WINDOW(l->data);
		if (!(wnck_window_is_skip_pager(window) ||
		      wnck_window_is_skip_tasklist(window)))
			window_list = g_list_append(window_list, l->data);
		l = l->next;
	}
	g_list_free(l);
	return window_list;
}

/* Returns a list of the windows on the active workspace */
static GList *get_workspace_window_list(WnckWorkspace *work_space,
					GList *wins)
{
	GList *window_list = NULL;
	while (wins)
	{
		WnckWindow *window = WNCK_WINDOW(wins->data);
		if (wnck_window_is_on_workspace(window,work_space))
			window_list = g_list_append(window_list,wins->data);
		wins = wins->next;
	}
	wins = g_list_first(wins);
	return window_list;
}

/* Returns the selected workspace from the pager widget */
static WnckWorkspace *get_selected_workspace(GdkEventButton *event)
{
	int space_number;
	
	space_number = workspace_at_point(WNCK_PAGER(pager),
					  event->x, event->y,
					  NULL, NULL);
	if (space_number == -1)
		return wnck_screen_get_active_workspace(screen);

	return wnck_screen_get_workspace(screen, space_number);
}

static gboolean button_event(GtkWidget *widget,
			     GdkEventButton *event,
			     gpointer user_data)
{
	if (event->button == 3)
	{
		WnckWorkspace *work_space;
		GtkWidget *menu_all;
		GtkWidget *menu;
		GtkWidget *item;
		GList	*wins_all;
		GList	*wins;
		gchar   *pos = NULL;

		if (socket)
		{
			pos = read_string(socket,
				gdk_atom_intern("_ROX_PANEL_MENU_POS",
						FALSE));
		}
		
		wins_all = get_window_list();
#ifdef HAVE_WNCK_CREATE_WINDOW_MENU
		menu_all = wnck_create_window_menu(wins_all);
#else
		menu_all = our_create_window_menu(wins_all);
#endif
		
		work_space = get_selected_workspace(event);
		wins = get_workspace_window_list(work_space,wins_all);
#ifdef HAVE_WNCK_CREATE_WINDOW_MENU
		menu = wnck_create_window_menu(wins);
#else
		menu = our_create_window_menu(wins);
#endif
		
		if (wins != NULL && g_list_length(wins) > 0)
		{
			item = gtk_menu_item_new();
			gtk_menu_append(GTK_MENU(menu), item);
			gtk_widget_show(item);
		}
		
		item = gtk_menu_item_new_with_label(_("All windows"));
		gtk_menu_item_set_submenu(GTK_MENU_ITEM(item), menu_all);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		item = gtk_menu_item_new();
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		item = gtk_menu_item_new_with_label(_("Help"));
		gtk_signal_connect(GTK_OBJECT(item), "activate",
				GTK_SIGNAL_FUNC(show_help), NULL);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		item = gtk_menu_item_new_with_label(_("Options"));
		gtk_signal_connect(GTK_OBJECT(item), "activate",
				GTK_SIGNAL_FUNC(rox_options_show), NULL);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		item = gtk_menu_item_new_with_label(_("Quit"));
		gtk_signal_connect(GTK_OBJECT(item), "activate",
				GTK_SIGNAL_FUNC(gtk_main_quit), NULL);
		gtk_menu_append(GTK_MENU(menu), item);
		gtk_widget_show(item);

		gtk_menu_popup(GTK_MENU(menu), NULL, NULL,
				position, pos,
				event->button, event->time);
		return 1;
	}

	return 0;
}

typedef enum {
	ORIENT_UNREAD,
	ORIENT_HORIZ,
	ORIENT_VERT,
	ORIENT_UNKNOWN
} Orient;

static Orient panel_get_orient(void)
{
	static Orient panel_orient = ORIENT_UNREAD;

	if (panel_orient == ORIENT_UNREAD)
	{
		char *pos;

		panel_orient = ORIENT_UNKNOWN;

		if (socket)
		{
			pos = read_string(socket,
				gdk_atom_intern("_ROX_PANEL_MENU_POS", FALSE));

			if (pos[0] == 'T' || pos[0] == 'B')
				panel_orient = ORIENT_HORIZ;
			else if (pos[0] == 'L' || pos[0] == 'R')
				panel_orient = ORIENT_VERT;
			g_free(pos);
		}

	}
	return panel_orient;
}

static void work_out_size(GtkWidget *widget, int width, int height,
		Orient orient, int *width_out, int *height_out)
{
	double aspect;
	int border_width;
	int spaces_per_row;
	int n_screens;
	WnckWorkspace *wspace;

	/* width and height will eventually store size of each workspace, so
	 * preserve actual allocation to start with */
	*width_out = width;
	*height_out = height;
	if (orient == ORIENT_UNKNOWN)
	{
		return;
	}
	gtk_widget_style_get(widget, "focus-line-width", &border_width, NULL);
	border_width *= 2;
	if (frame_type.int_value != 2)
		border_width += 2 * widget->style->xthickness;
	width -= border_width;
	height -= border_width;
	n_screens = wnck_screen_get_workspace_count(screen);
	spaces_per_row = (n_screens + n_rows.int_value - 1) / n_rows.int_value;
	wspace = wnck_screen_get_workspace(screen, 0);
	if (wspace)
	{
		aspect = (double) wnck_workspace_get_width(wspace) /
              (double) wnck_workspace_get_height(wspace);
	}
	else
	{
		aspect = (double) gdk_screen_width() / (double) gdk_screen_height();
	}

	if (orient == ORIENT_HORIZ)
	{
		/* Get height and width of each workspace */
		height -= (n_rows.int_value - 1);	/* Subtract single pixel dividers */
		height /= n_rows.int_value;
		width = (int) (height * aspect);
		/* Now calculate total width */
		*width_out = width * spaces_per_row + (spaces_per_row - 1) +
			border_width;
	}
	else /* (orient == ORIENT_VERT) */
	{
		/* Get width and height of each workspace;
		 * "rows" are now really columns */
		width -= (n_rows.int_value - 1);	/* Subtract single pixel dividers */
		width /= n_rows.int_value;
		height = (int) (width / aspect);
		/* Now calculate total height */
		*height_out = height * spaces_per_row + (spaces_per_row - 1) +
			border_width;
	}
}

static void set_size_request(GtkWidget *widget,
		int alloc_width, int alloc_height)
{
	int width, height;
	Orient orient;

	/* If not in a panel do nothing */
	if (!socket)
		return;

	orient = panel_get_orient();
	work_out_size(widget, alloc_width, alloc_height, orient, &width, &height);

	/* We're only using the panel's height as a guide for setting the
	 * widget's width (vice versa for vertical panels) and must stick to 8
	 * for our height allocation */
	if (orient == ORIENT_VERT)
	{
		if (height == alloc_height)
			return;
		width = 8;
	}
	else
	{
		if (width == alloc_width)
			return;
		height = 8;
	}
	gtk_widget_set_size_request(widget, width, height);
}

static void resize_handler(GtkWidget *widget, GtkAllocation *allocation,
        gpointer data)
{
	set_size_request(widget, allocation->width, allocation->height);
	(void) data;
}

static void pack_hierarchy(gboolean repack)
{
	g_object_ref(G_OBJECT(pager));
	switch (frame_type.int_value)
	{
		case 0:
			if (frame)
			{
				gtk_container_remove(GTK_CONTAINER(frame), pager);
				gtk_widget_destroy(frame);
				frame = NULL;
				repack = FALSE;
			}
			wnck_pager_set_shadow_type(WNCK_PAGER(pager), GTK_SHADOW_IN);
			if (!repack)
			{
				gtk_container_add(GTK_CONTAINER(toplevel), pager);  
				gtk_widget_show(pager);
			}
			break;
		case 1:
		case 2:
			if (!frame)
			{
				gtk_container_remove(GTK_CONTAINER(toplevel), pager);
				frame = gtk_frame_new(NULL);
				gtk_container_add(GTK_CONTAINER(toplevel), frame);
				gtk_container_add(GTK_CONTAINER(frame), pager);
			}
			if (frame_type.int_value == 1)
			{
				gtk_frame_set_shadow_type(GTK_FRAME(frame),
						GTK_SHADOW_NONE);
				wnck_pager_set_shadow_type(WNCK_PAGER(pager),
						GTK_SHADOW_IN);
			}
			else
			{
				gtk_frame_set_shadow_type(GTK_FRAME(frame),
						GTK_SHADOW_IN);
				wnck_pager_set_shadow_type(WNCK_PAGER(pager),
						GTK_SHADOW_NONE);
			}
			gtk_widget_show_all(frame);
			break;
	}
	g_object_unref(G_OBJECT(pager));
}

static void options_changed(void)
{
	if (n_rows.has_changed)
	{
		wnck_pager_set_n_rows(WNCK_PAGER(pager), n_rows.int_value);
		set_size_request(pager, pager->allocation.width,
				pager->allocation.height);
	}
	else if (frame_type.has_changed)
	{
		pack_hierarchy(TRUE);
	}
}

static void set_trans(void)
{
	struct stat info;
	guchar	*path;
	gchar	*lang2 = NULL;
	const char *lang;

	gtk_set_locale();

	rox_clear_translation();

	lang = getenv("LANG");
	if (!lang)
		return;
	/* Extract the language code from the locale name.
	 * language[_territory][.codeset][@modifier]
	 */
	if (lang[0] != '\0' && lang[1] != '\0'
	    && (lang[2] == '_' || lang[2] == '.' || lang[2] == '@'))
		lang2 = g_strndup((gchar *) lang, 2);

	path = g_strdup_printf("%s/Messages/%s.gmo", app_dir,
			lang2 ? lang2 : lang);
	g_free(lang2);

	if (stat(path, &info) == 0)
		rox_add_translations(path);
	g_free(path);
}

int main(int argc, char **argv)
{
	app_dir = g_strdup(getenv("APP_DIR"));
	if (!app_dir)
		g_error("APP_DIR environment variable was unset!\n"
			"Use the AppRun script to invoke " PROJECT "...\n");

	rox_init_with_domain( PROJECT, SITE, &argc, &argv );

	set_trans();

	rox_option_add_int(&n_rows, "n_rows", 2);

	rox_option_add_int(&frame_type, "frame_type", 2);

	rox_option_add_notify(options_changed);

	/* Send errors to stdout. Prevents bugs in OpenOffice flooding the
	 * error stream.
	 */
	dup2(1, 2);

	screen = wnck_screen_get_default();

	/* because the pager doesn't respond to signals at the moment */
	wnck_screen_force_update(screen);

	if (argc == 2)
	{
		gint32 xid;

		xid = atol(argv[1]);
		toplevel = gtk_plug_new(xid);
		/*gtk_widget_set_size_request(toplevel, 100, 20);*/
		socket = gdk_window_foreign_new(xid);
	}
	else
	{
		GdkPixbuf *pixbuf;
		gchar *file;

		toplevel = gtk_window_new(GTK_WINDOW_TOPLEVEL);
		gtk_window_stick(GTK_WINDOW(toplevel));
		gtk_window_set_title(GTK_WINDOW(toplevel), "Pager");

		file = g_build_filename(app_dir, ".DirIcon", NULL);
		pixbuf = gdk_pixbuf_new_from_file(file, NULL);
		if (pixbuf)
		{
			gtk_window_set_icon(GTK_WINDOW(toplevel), pixbuf);
			gdk_pixbuf_unref(pixbuf);
		}
		g_free(file);
	}

	g_signal_connect(toplevel, "destroy",
			 G_CALLBACK(gtk_main_quit), NULL);

	pager = wnck_pager_new(screen);
	wnck_pager_set_n_rows(WNCK_PAGER(pager), n_rows.int_value);
	if (panel_get_orient() == ORIENT_VERT)
		wnck_pager_set_orientation(WNCK_PAGER(pager),
					GTK_ORIENTATION_VERTICAL);

	pack_hierarchy(FALSE);

	g_signal_connect(pager, "button-press-event",
			G_CALLBACK(button_event), NULL);
	if (argc == 2)
	{
		gtk_widget_set_size_request(pager, 8, 8);
		g_signal_connect(pager, "size-allocate",
				G_CALLBACK(resize_handler), NULL);
	}

	gtk_widget_show_all(toplevel);

	gtk_main();

	return 0;
}

/* Taken from libwnck/pager.c */
#define POINT_IN_RECT(xcoord, ycoord, rect) \
 ((xcoord) >= (rect).x &&                   \
  (xcoord) <  ((rect).x + (rect).width) &&  \
  (ycoord) >= (rect).y &&                   \
  (ycoord) <  ((rect).y + (rect).height))

#define N_SCREEN_CONNECTIONS 11

static void
get_workspace_rect (WnckPager    *pager,
                    int           space,
                    GdkRectangle *rect)
{
  int hsize, vsize;
  int n_spaces;
  int spaces_per_row;
  GtkWidget *widget;
  int col, row;
  int focus_width;
  int nrows = n_rows.int_value;

  gtk_widget_style_get (GTK_WIDGET (pager),
			"focus-line-width", &focus_width,
			NULL);

  widget = GTK_WIDGET (pager);

  hsize = widget->allocation.width - 2 * focus_width;
  vsize = widget->allocation.height - 2 * focus_width;

  if (frame_type.int_value != 2)
    {
      hsize -= 2 * widget->style->xthickness;
      vsize -= 2 * widget->style->ythickness;
    }
  
  n_spaces = wnck_screen_get_workspace_count (wnck_screen_get_default ());

  g_return_if_fail(nrows > 0);
  spaces_per_row = (n_spaces + nrows - 1) / nrows;
  
  if (panel_get_orient() == ORIENT_VERT)
    {      
      rect->width = (hsize - (nrows - 1)) / nrows;
      rect->height = (vsize - (spaces_per_row - 1)) / spaces_per_row;

      col = space / spaces_per_row;
      row = space % spaces_per_row;
      rect->x = (rect->width + 1) * col; 
      rect->y = (rect->height + 1) * row;
      
      if (col == nrows - 1)
	rect->width = hsize - rect->x;
      
      if (row  == spaces_per_row - 1)
	rect->height = vsize - rect->y;
    }
  else
    {
      rect->width = (hsize - (spaces_per_row - 1)) / spaces_per_row;
      rect->height = (vsize - (nrows - 1)) / nrows;
      
      col = space % spaces_per_row;
      row = space / spaces_per_row;
      rect->x = (rect->width + 1) * col; 
      rect->y = (rect->height + 1) * row;

      if (col == spaces_per_row - 1)
	rect->width = hsize - rect->x;
      
      if (row  == nrows - 1)
	rect->height = vsize - rect->y;
    }

  rect->x += focus_width;
  rect->y += focus_width;

  /* if (pager->priv->shadow_type != GTK_SHADOW_NONE) */
    {
      rect->x += widget->style->xthickness;
      rect->y += widget->style->ythickness;
    }
}

static int
workspace_at_point (WnckPager *pager,
                    int        x,
                    int        y,
                    int       *viewport_x,
                    int       *viewport_y)
{
  GtkWidget *widget;
  int i;
  int n_spaces;
  int focus_width;
  int xthickness;
  int ythickness;

  widget = GTK_WIDGET (pager);

  gtk_widget_style_get (GTK_WIDGET (pager),
			"focus-line-width", &focus_width,
			NULL);

  xthickness = focus_width + widget->style->xthickness;
  ythickness = focus_width + widget->style->ythickness;

  n_spaces = wnck_screen_get_workspace_count (screen);
  
  i = 0;
  while (i < n_spaces)
    {
      GdkRectangle rect = {0};
      
      get_workspace_rect (pager, i, &rect);

      /* if (pager->priv->shadow_type != GTK_SHADOW_NONE) */
	{
	  /* If workspace is on the edge, pretend points on the frame
	   * belong to the workspace.
	   */

	  GtkWidget *widget = GTK_WIDGET (pager);

	  if (rect.x == xthickness)
	    {
	      rect.x = 0;
	      rect.width += xthickness;
	    }
	  if (rect.y == ythickness)
	    {
	      rect.y = 0;
	      rect.height += ythickness;
	    }
	  if (rect.y + rect.height == widget->allocation.height - ythickness)
	    {
	      rect.height += ythickness;
	    }
	  if (rect.x + rect.width == widget->allocation.width - xthickness)
	    {
	      rect.width += xthickness;
	    }
	}

      if (POINT_IN_RECT (x, y, rect))
        {
	  double width_ratio, height_ratio;
	  WnckWorkspace *space;

	  space = wnck_screen_get_workspace (screen, i);
          g_assert (space != NULL);

          /* Scale x, y mouse coords to corresponding screenwide viewport coords */
          
          width_ratio = (double) wnck_workspace_get_width (space) / (double) rect.width;
          height_ratio = (double) wnck_workspace_get_height (space) / (double) rect.height;

          if (viewport_x)
            *viewport_x = width_ratio * (x - rect.x);
          if (viewport_y)
            *viewport_y = height_ratio * (y - rect.y);

	  return i;
	}

      ++i;
    }

  return -1;
}
